const Gettext = imports.gettext;
const GLib    = imports.gi.GLib;
const Gio     = imports.gi.Gio;
const Gtk     = imports.gi.Gtk;
const Lang    = imports.lang;
const Me      = imports.misc.extensionUtils.getCurrentExtension();
const _       = imports.gettext.domain(Me.metadata['gettext-domain']).gettext;

const Ui = Me.imports.ui;

let settings = {};

function init() {
    Gettext.bindtextdomain(Me.metadata['gettext-domain'], Me.dir.get_child('locale').get_path());

    const GioSSS = Gio.SettingsSchemaSource;

    let schema = Me.metadata['settings-schema'];
    let source = GioSSS.new_from_directory(Me.dir.get_child('schemas').get_path(),
                                           GioSSS.get_default(),
                                           false);

    settings = new Gio.Settings({
        settings_schema : source.lookup(schema, true)
    });

    let set_strv = Lang.bind(settings, settings.set_strv);
    let set_boolean = Lang.bind(settings, settings.set_boolean);
    settings.set_strv = function(key, value) {
        set_strv(key, value);
        Gio.Settings.sync();
    };
    settings.set_boolean = function(key, value) {
        set_boolean(key, value);
        Gio.Settings.sync();
    };
    settings.set_strw = function(key, value) {
        settings.set_value(key, GLib.Variant.new('aas', value));
        Gio.Settings.sync();
    };
    settings.get_strw = function(key) {
        let aas  = [];
        for (let i = 0, outer = settings.get_value(key); i < outer.n_children(); ++i) {
            aas[i] = [];
            for (let j = 0, inner = outer.get_child_value(i); j < inner.n_children(); ++j) {
                aas[i][j] = inner.get_child_value(j).get_string().toString().split(',', 1)[0];
            }
        }

        return aas;
    };

    // show it in dConf editor
    workaroundWorkaround();
}

function getAccel(name) {
    return Gtk.accelerator_parse(settings.get_strv(name)[0]);
}

function unsetAccel(name) {
    let accels = settings.get_strv(name);
    accels[0] = '';
    settings.set_strv(name, accels);
}

function setAccel(name, key, mods) {
    let accels = settings.get_strv(name);
    accels[0] = Gtk.accelerator_name(key, mods);
    settings.set_strv(name, accels);
}

function connect() {
    let hooks = [];
    for (let i = 1; i < arguments.length - 1; ++i) {
        hooks.push(arguments[0].connect(arguments[i], arguments[arguments.length - 1]));
    }

    let actor = arguments[0];
    return function() {
        while(hooks.length > 0) {
            actor.disconnect(hooks.pop());
        }
    };
}

function override(obj, func, before) {
    let orig = obj[func];
    obj[func] = function(direct) {
        if (direct == true || before()) {
            orig.call(obj);
        }
    };
    return function() {
        obj[func] = orig;
    };
}

// Workaround #1

function getWorkaroundDelay() {
    return settings.get_int('workaround-delay');
}

function workaroundWorkaround() {
    settings.set_int('workaround-delay', getWorkaroundDelay());
}

// Tweak #1

function getCompactSearchResults() {
    return settings.get_boolean('compact-search-results');
}

function toggleCompactSearchResults() {
    settings.set_boolean('compact-search-results', !getCompactSearchResults());
}

function onSearchResultsStyleChange(callback) {
    return connect(settings, 'changed::compact-search-results', callback);
}

// Tweak #2

function getCompactApplications() {
    return settings.get_boolean('compact-applications');
}

function toggleCompactApplications() {
    settings.set_boolean('compact-applications', !getCompactApplications());
}

function onAppsStyleChange(callback) {
    return connect(settings, 'changed::compact-applications', callback);
}

// Tweak #3

function getOpaqueOverview() {
    return settings.get_boolean('opaque-overview');
}

function toggleOpaqueOverview() {
    settings.set_boolean('opaque-overview', !getOpaqueOverview());
}

function onOverviewStyleChange(callback) {
    return connect(settings, 'changed::opaque-overview', callback);
}

// Menus

function getMenus() {
    return settings.get_strw('menus');
}

function setMenus(menus) {
    settings.set_strw('menus', menus);
}

// Hidden apps

function getHiddenApps() {
    return settings.get_strv('hidden-apps');
}

function setHiddenApps(apps) {
    return settings.set_strv('hidden-apps', apps);
}

// Moved apps

function getMovedApps() {
    return settings.get_strv('moved-apps');
}

function setMovedApps(apps) {
    return settings.set_strv('moved-apps', apps);
}

function onUserMenuChange(callback) {
    return connect(settings, 'changed::moved-apps', callback);
}

function onAppsChange(callback) {
    return connect(settings,
                   'changed::menus',
                   'changed::hidden-apps',
                   'changed::moved-apps',
                   callback);
}

// Workaround #2

function getAppFilenames() {
    return settings.get_strv('hack-app-filenames');
}

function getAppNames() {
    return settings.get_strv('hack-app-names');
}

function getMenuIds() {
    return settings.get_strv('hack-menu-ids');
}

function getMenuNames() {
    return settings.get_strv('hack-menu-names');
}

function setHacks(appFilenames, appNames, menuIds, menuNames) {
    settings.set_strv('hack-app-filenames', appFilenames);
    settings.set_strv('hack-app-names', appNames);
    settings.set_strv('hack-menu-ids', menuIds);
    settings.set_strv('hack-menu-names', menuNames);
}

function onHacksHack(callback) {
    return connect(settings,
                   'changed::hack-app-filenames',
                   'changed::hack-app-names',
                   'changed::hack-menu-ids',
                   'changed::hack-menu-names',
                   callback);
}

// Keybindings

function onViewApps(callback) {
    let noFlags = imports.gi.Meta.KeyBindingFlags.NONE;
    global.display.add_keybinding('apps', settings, noFlags, callback);

    return function() {
        global.display.remove_keybinding('apps');
    };
}

function onSearch(callback){
    let noFlags = imports.gi.Meta.KeyBindingFlags.NONE;
    global.display.add_keybinding('search', settings, noFlags, callback);

    return function() {
        global.display.remove_keybinding('search');
    };
}

// Widgets

function buildKeysWidget() {
    return Ui.accelView(
        function(model, accel) {
            let all = {
                'apps'   : _('Show applications page'),
                'search' : _('Show search page')
            };

            for (let name in all) {
                let [ key, mods ] = getAccel(name);
                model.set(model.append(), [ 0, 1, 2, 3 ], [ name, all[name], mods, key ]);
            }

            function update(name, iter) {
                let [ key, mods ] = getAccel(name);
                model.set(iter, [ 2, 3 ], [ mods, key ]);
            }

            connect(accel, 'accel-edited', function(accel, path, key, mods) {
                let [ ok, iter ] = model.get_iter_from_string(path);
                if (ok) {
                    let name = model.get_value(iter, 0);
                    setAccel(name, key, mods);
                    update(name, iter);
                }
            });

            connect(accel, 'accel-cleared', function(accel, path) {
                let [ ok, iter ] = model.get_iter_from_string(path);
                if (ok) {
                    let name = model.get_value(iter, 0);
                    unsetAccel(name);
                    update(name, iter);
                }
            });
        }
    );
}

function buildMenusWidget() {
    let menuIds = getMenuIds();
    let menuNames = getMenuNames();

    function init(model, group) {
        let ids = group.filter(function(menu, i, a) {
            return menuIds.indexOf(menu) >= 0;
        }).map(function(menu, i, a) {
            return menu.replace(/^\//g, '');
        });
        let names = ids.map(function(id, i, a) {
            return menuNames[menuIds.indexOf(id)];
        });

        let parent = model.append(null);
        model.set(parent, [ 0, 1 ], [ '/' + ids.join(', '), names.join(', ') ]);

        ids.forEach(function(id, i, a) {
            model.set(model.append(parent), [ 0, 1 ], [ id, names[i] ]);
        });
    }

    function getTheRest() {
        let existing = getMenus().reduce(function(all, menus, i, a) {
            return all.concat(menus);
        }, []);
        return menuIds.filter(function(menu, i, a) {
            return existing.indexOf(menu) < 0;
        });
    }

    let disconnect;
    function reconnect(model) {
        disconnect = connect(model, 'row-deleted', function() {
            update(model);
        });
    }

    function update(model) {
        disconnect();

        function foreach(f) {
            for (let i = 0, [ ok, iter ] = model.get_iter_from_string('0'); ok;
                 [ ok, iter ] = model.get_iter_from_string('' + ++i)) f(iter);
        }

        foreach(function(iter) {
            if (!model.iter_has_child(iter)) {
                let id = model.get_value(iter, 0);
                if (id.length > 0 && id[0] == '/') {
                    model.remove(iter);
                }
            }
        });

        let menus = [];

        foreach(function(iter) {
            if (model.iter_has_child(iter)) {
                let ids = [];
                model.filter_new(model.get_path(iter)).foreach(
                    function(model, path, iter, data) {
                        let id = model.get_value(iter, 0);
                        if (id != null) {
                            ids.push(id);
                        }
                    }
                );
                let names = ids.map(function(id, i, a) {
                    return menuNames[menuIds.indexOf(id)];
                });
                model.set(iter, [ 0, 1 ], [ '/' + ids.join(', '), names.join(', ') ]);
                menus.push(ids);
            } else {
                let id = model.get_value(iter, 0);
                let name = model.get_value(iter, 1);
                model.set(iter, [ 0, 1 ], [ '/' + id, name ]);
                model.set(model.append(iter), [ 0, 1 ], [ id, name ]);
                menus.push([ id ]);
            }
        });

        menus.pop();
        setMenus(menus);

        imports.mainloop.timeout_add(getWorkaroundDelay(), function() {
            foreach(function(iter) {
                model.row_has_child_toggled(model.get_path(iter), iter);
            });
        });

        reconnect(model);
    }

    return Ui.dndView(
        function(model) { // init
            getMenus().forEach(function(group, i, a) {
                init(model, group);
            });
            init(model, getTheRest());

            reconnect(model);
        },
        function(model, src, dest, drop) { // forbid drop
            return src.get_indices()[0] == dest.get_indices()[0]
                && !model.get_iter_from_string('' + (src.get_indices()[0] + 1))[0]
                || src.get_depth() == dest.get_depth() && drop > Ui.DROP_AFTER
                || src.get_depth() < dest.get_depth();
        },
        function(model, path) { // forbid drag
            return path.get_depth() == 1
                && !model.get_iter_from_string('' + (path.get_indices()[0] + 1))[0];
        }
    );
}


function buildAppsWidget() {
    let filenames = getAppFilenames();
    let names = getAppNames();

    function init(model, get, row) {
        let parent = model.append(null);
        model.set(parent, [ 0, 1 ], row);

        get().forEach(function(app, i, a) {
            let index = filenames.indexOf(app);
            let iter;
            if (index < 0) {
                model.set(model.append(parent), [ 0, 1 ], [ app, app ]);
            } else {
                model.set(model.append(parent), [ 0, 1 ], [ app, names[index] ]);
            }
        });
    }

    function getAvailableApps() {
        let existing = getMovedApps().concat(getHiddenApps());
        return filenames.filter(function(app, i, a) {
            return existing.indexOf(app) < 0;
        });
    }

    function update(model, set, path) {
        let [ ok, iter ] = model.get_iter_from_string(path);
        if (ok) {
            let apps = [];
            model.filter_new(model.get_path(iter)).foreach(
                function(model, path, iter, data) {
                    let app = model.get_value(iter, 0);
                    if (app != null) {
                        apps.push(app);
                    }
                }
            );
            set(apps);
        }
    }

    return Ui.dndView(
        function(model) { // init
            init(model, getAvailableApps, [ 'available', _('Applications menu') ]);
            init(model, getMovedApps, [ 'moved', _('User menu') ]);
            init(model, getHiddenApps, [ 'hidden', _('Hidden') ]);

            connect(model, 'row-deleted', function() {
                update(model, setMovedApps, '1');
                update(model, setHiddenApps, '2');
            });
        },
        function(model, src, dest, drop) { // forbid drop
            return src.get_indices()[0] == dest.get_indices()[0] && src.get_indices()[0] != 1
                || dest.get_depth() == 1 && drop <= Ui.DROP_AFTER
                || dest.get_depth() == 2 && drop > Ui.DROP_AFTER;
        },
        function(model, path) { // forbid drag
            return path.get_depth() == 1;
        }
    );
}

function buildTweaksWidget() {
    return Ui.grid(
        Ui.switchOption(
            _('Remove horizontal spacing in applications view'),
            getCompactApplications,
            toggleCompactApplications
        ),
        Ui.switchOption(
            _('Remove vertical spacing in search results (including the applications view)'),
            getCompactSearchResults,
            toggleCompactSearchResults
        ),
        Ui.switchOption(
            _('Black overview (hide the desktop icons, if they exist)'),
            getOpaqueOverview,
            toggleOpaqueOverview
        )
    );
}

function buildPrefsWidget() {
    let all = {
        keys : {
            label  : _('Shortcuts'),
            widget : buildKeysWidget()
        },
        menus : {
            label  : _('Menus'),
            widget : buildMenusWidget()
        },
        apps : {
            label  : _('Applications'),
            widget : buildAppsWidget()
        },
        tweaks : {
            label  : _('Tweaks'),
            widget : buildTweaksWidget()
        }
    };

    let prefs = Ui.sidebarView(
        function(model) {
            for (let opt in all) {
                model.set(model.append(), [ 0, 1 ], [ opt, all[opt].label ]);
            }
        },
        function(opt) {
            return all[opt].widget;
        }
    );

    prefs.show_all();
    return prefs;
}
